import re

from fabric.api import cd, run, sudo
from fabric.contrib import files

from fabulaws.decorators import cached_property, uses_fabric
from fabulaws.ubuntu.packages.base import AptMixin


class PostgresMixin(AptMixin):
    """
    FabulAWS Ubuntu mixin that installs and configures PostgresSQL.
    """

    package_name = "postgresql"

    postgresql_packages = ["postgresql", "libpq-dev"]
    postgresql_tune = True
    postgresql_tune_type = "Web"
    postgresql_shmmax = 107374182400  # 100 GB
    postgresql_shmall = 26214400  # 100 GB / PAGE_SIZE (4096)
    # for help adjusting these settings, see:
    # http://wiki.postgresql.org/wiki/Tuning_Your_PostgreSQL_Server
    # http://wiki.postgresql.org/wiki/Number_Of_Database_Connections
    # http://thebuild.com/presentations/not-my-job-djangocon-us.pdf
    postgresql_settings = {
        # connections
        "max_connections": "80",  # _active_ connections are limited by pgbouncer
        # replication settings
        "wal_level": "hot_standby",
        "hot_standby": "on",
        "hot_standby_feedback": "on",
        "max_wal_senders": "3",
        # during client deletion 50 or more may be generated per minute; this allows an hour
        "wal_keep_segments": "3000",
        # resources - set these dynamically based on actual machine resources (see pg_resource_settings())
        # 'shared_buffers': '8GB',
        # 'work_mem': '750MB',
        # 'maintenance_work_mem': '1GB',
        # 'effective_cache_size': '48GB',
        # checkpoint settings
        "wal_buffers": "16MB",
        "checkpoint_completion_target": "0.9",
        "checkpoint_timeout": "10min",
        # if checkpoints are happening more often than the timeout, increase this up to 256
        "checkpoint_segments": "256",
        # logging
        "log_min_duration_statement": "500",
        "log_checkpoints": "on",
        "log_lock_waits": "on",
        "log_temp_files": "0",
        # write optimizations
        "commit_delay": "4000",  # delay each commit this many microseconds in case we can do a group commit
        "commit_siblings": "5",  # only delay if at least N transactions are in process
        # index usage optimizations
        # our DB servers have a lot of RAM and may tend to prefer Seq Scans if this is too high
        "random_page_cost": "2",
    }
    postgresql_networks = ["10.0.0.0/8", "172.16.0.0/12"]
    postgresql_disable_oom = False

    def __init__(self, *args, **kwargs):
        db_settings = kwargs.pop("db_settings", {}).copy()
        for key in [
            "postgresql_packages",
            "postgresql_tune",
            "postgresql_tune_type",
            "postgresql_shmmax",
            "postgresql_shmall",
            "postgresql_networks",
            "postgresql_disable_oom",
        ]:
            if key in db_settings:
                setattr(self, key, db_settings.pop(key))

        # Override individual default settings with whatever settings the project has specified.
        self.postgresql_settings = self.postgresql_settings.copy()
        self.postgresql_settings.update(db_settings.pop("postgresql_settings", {}))
        self.pg_pw_encryption = self.postgresql_settings.get(
            "password_encryption", "md5"
        )
        if db_settings:
            # There were keys we did not recognize; complain rather than let the
            # user think we're applying setttings that we're not.
            raise ValueError(
                "Unrecognized keys in 'db_settings': %s" % ", ".join(db_settings.keys())
            )

        super(PostgresMixin, self).__init__(*args, **kwargs)

    @property
    def pgpass(self):
        return "/var/lib/postgresql/.pgpass"

    @cached_property()
    @uses_fabric
    def pg_version(self):
        version = run("pg_config --version")
        version = re.findall(r"(\d+\.\d+)\.?", version)[0]
        if float(version) >= 10:
            # For Postgres 10 and newer, return just the major version number (used in file paths)
            return version.split(".")[0]
        else:
            # Otherwise, return the full version, e.g., "9.5"
            return version

    @cached_property()
    @uses_fabric
    def pg_conf(self):
        return "/etc/postgresql/{0}/main/postgresql.conf".format(self.pg_version)

    @cached_property()
    @uses_fabric
    def pg_data(self):
        return "/var/lib/postgresql/{0}/main".format(self.pg_version)

    @cached_property()
    @uses_fabric
    def pg_hba(self):
        return "/etc/postgresql/{0}/main/pg_hba.conf".format(self.pg_version)

    @cached_property()
    @uses_fabric
    def pg_bin(self):
        return "/usr/lib/postgresql/{0}/bin".format(self.pg_version)

    @uses_fabric
    def sed(self, before, after, file_):
        # fabric doesn't properly escape single quotes for sed commands, so run sed manually instead
        sudo(
            'sed -i.bak -r -e "s/{before}/{after}/g" {file_}'
            "".format(before=before, after=after, file_=file_)
        )

    @uses_fabric
    def pg_set_str(self, setting, value):
        self.sed(
            "^#? ?{setting} = '.*'".format(setting=setting),
            "{setting} = '{value}'".format(setting=setting, value=value),
            self.pg_conf,
        )

    @uses_fabric
    def pg_set(self, setting, value):
        self.sed(
            r"^#? ?{setting} = \S+".format(setting=setting),
            "{setting} = {value}".format(setting=setting, value=value),
            self.pg_conf,
        )

    @uses_fabric
    def pg_cmd(self, action, fail=True):
        """Run the specified action (e.g., start, stop, restart) on the postgresql server."""

        if fail or files.exists("/etc/init/postgresql.conf"):
            sudo("service postgresql %s" % action)

    @uses_fabric
    def pg_resource_settings(self):
        """
        Calculate a few resource settings dynamically. Will be overridden by
        pgtune or manually specified settings, if any.
        """
        mem = self.server_memory
        max_connections = int(self.postgresql_settings["max_connections"])
        # pgtune isn't available anymore as of Ubuntu 16.04, so calculate a few
        # basic resources dynamically just in case
        return {
            # 25% of available RAM, up to 8GB
            "shared_buffers": "%sMB" % int(min(mem * 0.25, 8096)),
            # (2*RAM)/max_connections
            "work_mem": "%sMB" % int((mem * 2) / max_connections),
            # RAM/16 up to 1GB; high values aren't that helpful
            "maintenance_work_mem": "%sMB" % int(min(mem / 16, 1024)),
            # between 50-75%, should equal free + cached values in `top`
            "effective_cache_size": "%sMB" % int(mem * 0.7),
        }

    @uses_fabric
    def pg_tune_config(self, restart=True):
        """Tune the postgresql configuration using pgtune"""

        self.install_packages(["pgtune"])
        old = "%s.bak" % self.pg_conf
        new = "%s.new" % self.pg_conf
        db_type = self.postgresql_tune_type
        conns = ""
        if "max_connections" in self.postgresql_settings:
            conns = "-c %s" % self.postgresql_settings["max_connections"]
        sudo("pgtune -T %s -i %s -o %s %s" % (db_type, self.pg_conf, new, conns))
        sudo("mv %s %s" % (self.pg_conf, old))
        sudo("mv %s %s" % (new, self.pg_conf))
        if restart:
            self.pg_cmd("restart")

    @uses_fabric
    def pg_set_sysctl_params(self, restart=True):
        sudo("sysctl -w kernel.shmmax=%s" % self.postgresql_shmmax)
        files.append(
            "/etc/sysctl.conf",
            "kernel.shmmax=%s" "" % self.postgresql_shmmax,
            use_sudo=True,
        )
        sudo("sysctl -w kernel.shmall=%s" % self.postgresql_shmall)
        files.append(
            "/etc/sysctl.conf",
            "kernel.shmall=%s" "" % self.postgresql_shmall,
            use_sudo=True,
        )
        if self.postgresql_disable_oom:
            sudo("sysctl -w vm.overcommit_memory=2")
            files.append("/etc/sysctl.conf", "vm.overcommit_memory=2", use_sudo=True)
        if restart:
            self.pg_cmd("restart")

    @uses_fabric
    def pg_replace_pw_encryption(self, before="scram-sha-256", after="md5"):
        files.sed(self.pg_hba, before=before, after=after, flags="i", use_sudo=True)

    @uses_fabric
    def pg_allow_from(self, ip_ranges, restart=True):
        """Allow external connections from the given IP range."""
        self.pg_set_str("listen_addresses", "*")
        files.uncomment(self.pg_hba, "local +replication", use_sudo=True)
        for ip_range in ip_ranges:
            hostssl_line = (
                f"hostssl    all    all    {ip_range} {self.pg_pw_encryption}"
            )
            files.append(self.pg_hba, hostssl_line, use_sudo=True)
        if restart:
            self.pg_cmd("restart")

    def pg_update_settings(self, settings, restart=True):
        """Update the specified settings according to the given dictionary."""

        for k, v in settings.items():
            self.pg_set(k, v)
        if restart:
            self.pg_cmd("restart")

    @uses_fabric
    def pg_allow_replication(self, user, password, ip_ranges, restart=True):
        """Creates a user for replication and enables replication in pg_hba.conf."""

        # XXX: does not support differing primary/replica pg versions
        self.create_db_user(user, password, replication=True)
        files.uncomment(self.pg_hba, "local +replication", use_sudo=True)
        for ip_range in ip_ranges:
            hostssl_line = (
                f"hostssl    replication    all   {ip_range}   {self.pg_pw_encryption}"
            )
            files.append(self.pg_hba, hostssl_line, use_sudo=True)
        if restart:
            sudo("service postgresql restart")

    @uses_fabric
    def pg_copy_master(self, master_db, user, password):
        """Replaces this database host with a copy of the data at master_host."""

        self.pg_cmd("stop")
        sudo("rm -rf {0}".format(self.pg_data))
        pgpass_line = ":".join(
            [master_db.internal_ip, "*", "replication", user, password]
        )
        sudo(
            'echo "{line}" > {file_}' "".format(file_=self.pgpass, line=pgpass_line),
            user="postgres",
        )
        sudo("chmod 600 {0}".format(self.pgpass), user="postgres")
        sudo(
            "{pg_bin}/pg_basebackup -X stream -D {pg_data} -P -h {host} -U {user}"
            "".format(
                pg_bin=self.pg_bin,
                pg_data=self.pg_data,
                host=master_db.internal_ip,
                user=user,
            ),
            user="postgres",
        )
        with cd(self.pg_data):
            signal = "standby.signal"
            sudo("touch {file_}".format(file_=signal), user="postgres")
            self.pg_set_str(
                "primary_conninfo",
                "host={host} user={user} password={password}"
                "".format(host=master_db.internal_ip, user=user, password=password),
            )
            sudo("ln -s /etc/ssl/certs/ssl-cert-snakeoil.pem server.crt")
            sudo("ln -s /etc/ssl/private/ssl-cert-snakeoil.key server.key")
        self.pg_cmd("start")

    @uses_fabric
    def pg_promote(self):
        sudo(
            "{0}/pg_ctl -D {1} promote".format(self.pg_bin, self.pg_data),
            user="postgres",
        )

    def bind_app_directories(self, *args, **kwargs):
        # make sure we stop first in case we're being moved to a secure directory
        self.pg_cmd("stop", fail=False)
        super(PostgresMixin, self).bind_app_directories(*args, **kwargs)
        self.pg_cmd("start", fail=False)

    def setup(self):
        """Postgres mixin"""

        super(PostgresMixin, self).setup()
        if "max_connections" in self.postgresql_settings:
            self.pg_update_settings(self.pg_resource_settings(), restart=False)
        if self.postgresql_tune:
            self.pg_tune_config(restart=False)
        self.pg_set_sysctl_params(restart=False)
        self.pg_replace_pw_encryption()
        self.pg_allow_from(self.postgresql_networks, restart=False)
        self.pg_update_settings(self.postgresql_settings, restart=False)
        self.pg_cmd("restart")

    @uses_fabric
    def sql(self, sql):
        sudo('psql -c "%s"' % sql, user="postgres")

    @uses_fabric
    def create_db_user(self, username, password=None, **kwargs):
        """Create a database user."""

        defaults = {"login": True}
        defaults.update(kwargs)
        options = [k.upper() for k in defaults if k]
        options.extend(["NO" + k.upper() for k in defaults if not k])
        sql = "CREATE ROLE {name} {options}".format(
            name=username, options=" ".join(options)
        )
        if password is not None:
            sql += " PASSWORD '%s'" % password
        self.sql(sql)

    @uses_fabric
    def change_db_user_password(self, username, password):
        """Change a db user's password."""

        self.sql("ALTER USER %s WITH PASSWORD '%s'" % (username, password))

    @uses_fabric
    def create_db(self, name, owner=None, encoding="UTF-8"):
        """Create a Postgres database."""

        flags = ""
        if encoding:
            flags = "-E %s" % encoding
        if owner:
            flags = "%s -O %s" % (flags, owner)
        sudo("createdb %s %s" % (flags, name), user="postgres")



import json
import os
import tempfile

from fabric.api import env, local, put, run, sudo
from fabric.operations import _prefix_commands, _prefix_env_vars

__all__ = [
    "sshagent_run",
    "call_python",
    "ec2_hostnames",
    "ec2_instances",
    "answer_sudo",
]


def sshagent_run(cmd, user=None):
    """
    Helper function.
    Runs a command with SSH agent forwarding enabled.

    Note:: Fabric (and paramiko) can't forward your SSH agent.
    This helper uses your system's ssh to do so.
    """
    # Handle context manager modifications
    wrapped_cmd = _prefix_commands(_prefix_env_vars(cmd), "remote")
    if user is None:
        user = env.user
    opts = ["-A", "-o ServerAliveInterval=60"]
    if env.disable_known_hosts:
        opts += ["-o StrictHostKeyChecking=no", "-o UserKnownHostsFile=/dev/null"]
    try:
        host, port = env.host_string.split(":")
        opts.append("-p %s" % port)
    except ValueError:
        host = env.host_string
    opts = " ".join(opts)
    return local("ssh %s %s@%s '%s'" % (opts, user, host, wrapped_cmd))


def call_python(method, *args):
    """
    Call the given ``method'' with the given ``args'' in Python on the
    remote server.  ``method'' should be the full Python path to the
    method.  Only JSON-serializable arguments and return values are
    supported.
    """
    module = ".".join(method.split(".")[:-1])
    args = json.dumps(args)[1:-1]
    output = run(
        "/usr/bin/env python3 -c 'import json, {module};"
        "print(json.dumps({method}({args})))'"
        "".format(module=module, method=method, args=args)
    )
    return json.loads(output)


def ec2_hostnames(*args, **kwargs):
    """
    Returns a list of hostnames for the specified filters.
    """
    from fabulaws.ec2 import EC2Service

    return EC2Service().public_dns(*args, **kwargs)


def ec2_instances(*args, **kwargs):
    """
    Returns a list of instances for the specified filters.
    """
    from fabulaws.ec2 import EC2Service

    return EC2Service().instances(*args, **kwargs)


def answer_sudo(cmd, *args, **kwargs):
    """
    Answers questions presented via standard out according to the
    given answers.  The ``answers`` keyword argument should be a list of
    (question, answer) pairs.  The questions are in regular expression format,
    so ensure that any special characters are appropriately escaped.
    """
    answers = kwargs.pop("answers", [])
    if answers:
        # use shared memory rather than potentially writing passwords to the
        # most likely unencrypted disk
        tmpdir = os.path.exists("/dev/shm") and "/dev/shm" or None
        script = tempfile.NamedTemporaryFile(dir=tmpdir)
        script.writelines(
            [
                "import pexpect, sys\n",
                "child = pexpect.spawn('{0}')\n".format(cmd),
                # the following line may be enabled for debugging purposes, but
                # will spill passphrases in plain text in log files
                # "child.logfile = sys.stdout\n"
            ]
        )
        for question, answer in answers:
            script.writelines(
                [
                    "child.expect('{0}')\n".format(question),
                    "child.sendline('{0}')\n".format(answer),
                ]
            )
        # this is important, otherwise pexect will kill the process before it's
        # finished
        script.writelines(["child.wait()\n"])
        script.flush()
        remote_file = "/dev/shm/{0}".format(os.path.basename(script.name))
        put(script.name, remote_file, mirror_local_mode=True)
        cmd = "python {0}".format(remote_file)
    result = sudo(cmd, *args, **kwargs)
    if answers:
        run("rm {0}".format(remote_file))
    return result



import logging
import os

from fabric.api import env, settings, sudo
from fabric.contrib import files

from fabulaws.decorators import uses_fabric
from fabulaws.ubuntu.instances import UbuntuInstance
from fabulaws.ubuntu.packages.fail2ban import Fail2banMixin
from fabulaws.ubuntu.packages.memcached import MemcachedMixin
from fabulaws.ubuntu.packages.postgres import PostgresMixin
from fabulaws.ubuntu.packages.python import PythonMixin
from fabulaws.ubuntu.packages.rabbitmq import RabbitMqMixin
from fabulaws.ubuntu.packages.redis import RedisMixin

logger = logging.getLogger(__name__)

__all__ = [
    "CacheInstance",
    "DbPrimaryInstance",
    "DbReplicaInstance",
    "WorkerInstance",
    "WebInstance",
    "CombinedInstance",
]


class FirewallMixin(Fail2banMixin):
    pass


class BaseInstance(FirewallMixin, UbuntuInstance):
    deployment_dir = os.path.dirname(__file__)
    project_root = os.path.dirname(deployment_dir)

    def __init__(self, *args, **kwargs):
        for key in [
            "ami",
            "key_prefix",
            "admin_groups",
            "run_upgrade",
            "app_root",
            "deploy_user_home",
            "fs_type",
            "fs_encrypt",
            "ubuntu_mirror",
            "swap_multiplier",
            "instance_type",
            "deploy_user",
            "volume_size",
            "volume_type",
            "security_groups",
            "ebs_encrypt",
        ]:
            setattr(self, key, kwargs.pop(key, ""))
        if "terminate" not in kwargs:
            kwargs["terminate"] = False
        passwd = getattr(env, "luks_passphrase", None)
        self.volume_info = [
            ("/dev/sdf", self.app_root, self.volume_size, self.volume_type, passwd)
        ]
        self.default_swap_file = "%s/swapfile" % self.app_root
        super(BaseInstance, self).__init__(*args, **kwargs)
        self.app_dirs = ["/tmp", self.deploy_user_home]  # mixins add other dirs

    def _get_users(self):
        """
        Returns a list of tuples of (username, key_file_path).
        """
        users_dir = env.ssh_keys
        users = [
            (n, os.path.join(users_dir, n))
            for n in os.listdir(users_dir)
            if not n.endswith(".gecos")
        ]  # skip files ending in '.gecos'
        return users

    def _add_swap(self, path):
        sudo("mkswap -f {0}".format(path))
        with settings(warn_only=True):
            # sometimes mkswap seems to 'mount' the swap partition
            # automatically, so this command will fail
            sudo("swapon {0}".format(path))
        files.append("/etc/fstab", "{0} none swap sw 0 0".format(path), use_sudo=True)

    @uses_fabric
    def setup_swap(self):
        """Sets up swap partition"""
        swap_mb = self.server_memory * self.swap_multiplier
        devs = ["/dev/xvdb", "/dev/xvdc"]
        for dev in devs:
            print("attempting swap creation on {0}...".format(dev))
            if swap_mb <= 0:
                print("no additional swap needed; skipping")
                break
            if not files.exists(dev):
                print("no such device {0}; skipping".format(dev))
                continue
            size = sudo("blockdev --getsize64 {0}".format(dev))
            try:
                size = int(size) / 1024 / 1024
            except ValueError:
                print("no size found for {0}".format(dev))
                continue
            # decrement size regardless of whether or not we create swap; if
            # crypt device exists, assume we created it previously and don't
            # create even more swap
            swap_mb -= size
            with settings(warn_only=True):
                sudo("umount {0}".format(dev))
            crypt_name = "cryptswap-{0}".format(dev.split("/")[-1])
            if files.exists("/dev/mapper/{0}".format(crypt_name)):
                print("crypt device {0} already exists; skipping".format(crypt_name))
                continue
            sudo("cryptsetup -d /dev/urandom create {0} {1}".format(crypt_name, dev))
            files.append(
                "/etc/crypttab",
                "{0} {1} /dev/urandom swap".format(crypt_name, dev),
                use_sudo=True,
            )
            files.comment("/etc/fstab", dev, use_sudo=True)
            self._add_swap("/dev/mapper/{0}".format(crypt_name))
        if swap_mb > 0:
            if not files.exists(self.default_swap_file):
                sudo("fallocate -l {0}M {1}".format(swap_mb, self.default_swap_file))
                sudo("chown root:root {0}".format(self.default_swap_file))
                sudo("chmod 600 {0}".format(self.default_swap_file))
                self._add_swap(self.default_swap_file)
            else:
                print(
                    "swap file {0} already exists; skipping".format(
                        self.default_swap_file
                    )
                )
        # remove old swap at the end in case it's already in use (otherwise swapoff might fail)
        old_swap = ["/dev/xvda3"]
        for swap in old_swap:
            if files.contains("/proc/swaps", swap):
                sudo("swapoff {0}".format(swap))
                files.comment("/etc/fstab", swap, use_sudo=True)
                with settings(warn_only=True):
                    sudo("dd if=/dev/zero of={0} bs=1M".format(swap))

    @uses_fabric
    def setup_sudoers(self):
        """
        Creates the sudoers file on the server, based on the supplied template.
        """
        sudoers_file = os.path.join(self.deployment_dir, "templates", "sudoers")
        files.upload_template(
            sudoers_file, "/etc/sudoers.new", backup=False, use_sudo=True, mode=0o440
        )
        sudo("chown root:root /etc/sudoers.new")
        sudo("mv /etc/sudoers.new /etc/sudoers")

    def reset_authentication(self):
        """
        Delete's the 'ubuntu' user in the AMI once it's no longer needed.
        """
        user = self.user  # save user before it's overwritten by super() method
        super(BaseInstance, self).reset_authentication()
        if user == "ubuntu":
            with self:
                # disable the 'ubuntu' user so it can no longer log in
                sudo("usermod --expiredate 1 {}".format(user))

    @uses_fabric
    def create_deployer(self):
        """
        Creates a deployment user with a directory for Apache configurations.
        """
        user = self.deploy_user
        sudo("useradd -d {0} -m -s /bin/bash {1}".format(self.deploy_user_home, user))
        sudo("mkdir {0}/.ssh".format(self.deploy_user_home), user=user)

    @uses_fabric
    def update_deployer_keys(self):
        """
        Replaces deployer keys with the current sysadmin users keys.
        """
        user = self.deploy_user
        file_ = "{0}/.ssh/authorized_keys2".format(self.deploy_user_home)
        if files.exists(file_):
            sudo("rm {0}".format(file_), user=user)
        sudo("touch {0}".format(file_), user=user)
        for _, key_file in self._get_users():
            files.append(file_, open(key_file).read().strip(), use_sudo=True)

    def setup(self):
        """
        Creates sysadmin users and secures the required directories.
        """
        super(BaseInstance, self).setup()
        self.instance.modify_attribute(
            "blockDeviceMapping", ["%s=true" % v[0] for v in self.volume_info]
        )
        self.create_users(self._get_users())
        self.setup_sudoers()
        # needed for SSH agent forwarding during replication setup:
        self.reset_authentication()
        self.create_deployer()
        self.update_deployer_keys()
        self.bind_app_directories(self.app_dirs, self.app_root)
        self.setup_swap()  # after app (potentially secure) partition is created
        self.upgrade_packages()
        self.install_packages(["ntp"])  # keep date current


class SessionMixin(RedisMixin):
    """Mixin that creates a session store using Redis."""

    redis_bind = ""  # allow connections on all interfaces

    def __init__(self, *args, **kwargs):
        super(SessionMixin, self).__init__(*args, **kwargs)
        self.app_dirs.append("/var/lib/redis")


class CacheMixin(MemcachedMixin):
    """Mixin that creates a cache using Memcached."""

    memcached_bind = ""  # allow connections on all interfaces
    memcached_connections = 10000
    memcached_ulimit = 20000
    memcached_threads = 5

    def __init__(self, *args, **kwargs):
        super(CacheMixin, self).__init__(*args, **kwargs)

    @property
    @uses_fabric
    def memcached_memory(self):
        """Returns half of total server memory, in MB"""
        return self.server_memory / 2


class QueueMixin(RabbitMqMixin):
    """Mixin that creates a RabbitMQ user and host based on the Fabric env."""

    def __init__(self, *args, **kwargs):
        super(QueueMixin, self).__init__(*args, **kwargs)
        self.app_dirs.append("/var/lib/rabbitmq")

    def setup(self):
        """Create the RabbitMQ user and vhost."""

        super(QueueMixin, self).setup()
        self.create_mq_user(env.deploy_user, env.broker_password)
        self.create_mq_vhost(env.vhost)
        self.set_mq_vhost_permissions(env.vhost, env.deploy_user, '".*" ".*" ".*"')


class DbMixin(PostgresMixin):
    """Mixin that creates a database based on the Fabric env."""

    def __init__(self, *args, **kwargs):
        kwargs["db_settings"] = env.db_settings
        super(DbMixin, self).__init__(*args, **kwargs)
        self.app_dirs.append("/var/lib/postgresql")

    def setup(self):
        """Create the Postgres user and database based on the Fabric env."""

        super(DbMixin, self).setup()
        self.pg_allow_replication(
            "%s_repl" % env.database_user,
            env.database_password,
            self.postgresql_networks,
        )


class DbPrimaryMixin(DbMixin):
    """Mixin that creates a database based on the Fabric env."""

    def setup(self):
        """Create the Postgres user and database based on the Fabric env."""

        super(DbPrimaryMixin, self).setup()
        self.create_db_user(env.database_user, password=env.database_password)
        self.create_db(env.database_name, owner=env.database_user)


class DbReplicaMixin(DbMixin):
    """Mixin that creates a database based on the Fabric env."""

    def setup(self):
        """Create the Postgres user and database based on the Fabric env."""

        super(DbReplicaMixin, self).setup()
        if len(env.servers["db-primary"]) > 0:
            primary = env.servers["db-primary"][0]
            self.pg_copy_master(
                primary, "%s_repl" % env.database_user, env.database_password
            )


class AppMixin(PythonMixin):
    """
    Mixin that installs the Python application dependencies, including the appropriate
    versions of Python, PIP, and virtualenv.
    """

    @uses_fabric
    def install_less_and_yuglify(self):
        """
        Adds apt repo for node.js and NPM, installs NPM, and uses NPM to globally
        install the ``lessc`` binary used by django_compressor and pipeline
        and the ``yuglify`` binary used by pipeline.
        """

        node_version = getattr(env, "node_version", "6.x")
        sudo("curl -sL https://deb.nodesource.com/setup_%s | bash -" % node_version)
        sudo("export DEBIAN_FRONTEND=noninteractive ; apt-get -qq -y install nodejs")
        less_version = getattr(env, "less_version", "1.3.3")
        sudo("npm install -g less@%s" % less_version)
        sudo("npm install -g yuglify")

    @uses_fabric
    def install_system_packages(self):
        """Installs the system packages specified in the environment."""

        # Install required system packages for deployment, plus some extras
        packages = set(env.app_server_packages) | set(
            ["supervisor", "pgbouncer", "stunnel4"]
        )
        self.install_packages(packages)
        # supervisord doesn't start automatically on all Ubuntu versions,
        # so make sure it's started here
        sudo("service supervisor restart")

    @uses_fabric
    def create_webserver_user(self):
        """Create a user for gunicorn, celery, etc."""

        if env.webserver_user != env.deploy_user:  # deploy_user already exists
            sudo("useradd --system %(webserver_user)s" % env)

    def setup(self):
        """
        Creates necessary directories, installs required packages, and copies
        the required SSH keys to the server.
        """

        super(AppMixin, self).setup()
        self.install_less_and_yuglify()
        self.install_system_packages()
        self.create_webserver_user()


class WorkerMixin(AppMixin):
    """Mixin that creates a web application server."""

    swap_multiplier = 4  # create 4x the server's RAM in swap

    def __init__(self, *args, **kwargs):
        super(WorkerMixin, self).__init__(*args, **kwargs)


class WebMixin(AppMixin):
    """Mixin that creates a web application server."""

    swap_multiplier = 0  # no swap on the web servers (unhealthy servers will be replaced by Auto Scaling)

    def __init__(self, *args, **kwargs):
        super(WebMixin, self).__init__(*args, **kwargs)

    def setup(self):
        """Installs nginx."""

        super(WebMixin, self).setup()
        self.install_packages(["nginx"])


class CacheInstance(QueueMixin, CacheMixin, SessionMixin, BaseInstance):
    pass


class DbPrimaryInstance(DbPrimaryMixin, BaseInstance):
    pass


class DbReplicaInstance(DbReplicaMixin, BaseInstance):
    pass


class WorkerInstance(WorkerMixin, BaseInstance):
    pass


class WebInstance(WebMixin, BaseInstance):
    pass


class CombinedInstance(DbPrimaryMixin, WebMixin, BaseInstance):
    pass



from fabric.api import sudo
from fabric.contrib import files

from fabulaws.decorators import uses_fabric
from fabulaws.ubuntu.packages.base import AptMixin


class RedisMixin(AptMixin):
    """
    FabulAWS Ubuntu mixin that installs and configures Redis.
    """

    package_name = "redis"
    redis_packages = ["redis-server"]
    redis_bind = "127.0.0.1"  # set to '' to bind to all interfaces
    redis_loglevel = "notice"
    redis_keepalive = 60
    redis_conf = "/etc/redis/redis.conf"
    redis_bind_pattern = r"#?\s*bind 127\.0\.0\.1"
    redis_protected_mode_pattern = r"#?\s*protected-mode \w+"
    redis_loglevel_pattern = r"#?\s*loglevel \w+"
    redis_keepalive_pattern = r"#?\s*tcp-keepalive \w+"

    @uses_fabric
    def redis_service(self, cmd):
        sudo("service redis-server {0}".format(cmd), pty=False)  # must pass pty=False

    @uses_fabric
    def redis_configure(self, bind=None, loglevel=None, keepalive=None):
        if bind is None:
            bind = self.redis_bind
        if loglevel is None:
            loglevel = self.redis_loglevel
        if keepalive is None:
            keepalive = self.redis_keepalive
        if bind == "":  # all interfaces
            files.comment(self.redis_conf, self.redis_bind_pattern, use_sudo=True)
            files.sed(
                self.redis_conf,
                self.redis_protected_mode_pattern,
                "protected-mode no",
                use_sudo=True,
            )
        else:
            files.sed(self.redis_conf, self.redis_bind_pattern, bind, use_sudo=True)
        loglevel = "loglevel {0}".format(loglevel)
        files.sed(self.redis_conf, self.redis_loglevel_pattern, loglevel, use_sudo=True)
        keepalive = "tcp-keepalive {}".format(keepalive)
        files.sed(
            self.redis_conf, self.redis_keepalive_pattern, keepalive, use_sudo=True
        )
        self.redis_service("restart")

    def bind_app_directories(self, *args, **kwargs):
        # make sure we stop first in case we're being moved to a secure directory
        self.redis_service("stop")
        super(RedisMixin, self).bind_app_directories(*args, **kwargs)
        self.redis_service("start")

    def setup(self):
        """Redis mixin"""

        super(RedisMixin, self).setup()
        self.redis_configure()



import time
from functools import wraps

from fabulaws.ec2 import EC2Instance

__all__ = ["cached_property", "uses_fabric"]


# copied from http://wiki.python.org/moin/PythonDecoratorLibrary#Cached_Properties
# (c) 2011 Christopher Arndt, MIT License
class cached_property(object):
    """Decorator for read-only properties evaluated only once within TTL period.

    It can be used to created a cached property like this::

        import random

        # the class containing the property must be a new-style class
        class MyClass(object):
            # create property whose value is cached for ten minutes
            @cached_property(ttl=600)
            def randint(self):
                # will only be evaluated every 10 min. at maximum.
                return random.randint(0, 100)

    The value is cached  in the '_cache' attribute of the object instance that
    has the property getter method wrapped by this decorator. The '_cache'
    attribute value is a dictionary which has a key for every property of the
    object which is wrapped by this decorator. Each entry in the cache is
    created only when the property is accessed for the first time and is a
    two-element tuple with the last computed property value and the last time
    it was updated in seconds since the epoch.

    The default time-to-live (TTL) is 300 seconds (5 minutes). Set the TTL to
    zero for the cached value to never expire.

    To expire a cached property value manually just do::

        del instance._cache[<property name>]

    """

    def __init__(self, ttl=300):
        self.ttl = ttl

    def __call__(self, fget, doc=None):
        self.fget = fget
        self.__doc__ = doc or fget.__doc__
        self.__name__ = fget.__name__
        self.__module__ = fget.__module__
        return self

    def __get__(self, inst, owner):
        now = time.time()
        try:
            value, last_update = inst._cache[self.__name__]
            if self.ttl > 0 and now - last_update > self.ttl:
                raise AttributeError
        except (KeyError, AttributeError):
            value = self.fget(inst)
            try:
                cache = inst._cache
            except AttributeError:
                cache = inst._cache = {}
            cache[self.__name__] = (value, now)
        return value


def uses_fabric(f):
    @wraps(f)
    def wrapper(self, *args, **kwds):
        if not isinstance(self, EC2Instance):
            raise ValueError(
                "@uses_fabric can only wrap methods on " "ECInstance classes"
            )
        with self:
            return f(self, *args, **kwds)

    return wrapper



import logging
from decimal import Decimal

from fabric.api import settings, sudo
from fabric.contrib import files

from fabulaws.decorators import uses_fabric

logger = logging.getLogger(__name__)


class BaseAptMixin(object):
    """
    FabulAWS mixin for manipulating Debian-based packages and their
    repositories
    """

    def _read_lines_from_file(self, file_name):
        with open(file_name) as f:
            packages = f.readlines()
        return [x.strip("\n\r") for x in packages]

    @uses_fabric
    def install_packages(self, packages):
        """Install apt packages from a list."""

        sudo(
            "export DEBIAN_FRONTEND=noninteractive ; apt-get -qq -y install %s"
            % " ".join(packages)
        )

    @uses_fabric
    def install_packages_from_file(self, file_name):
        """Install apt packages from a file list."""
        self.install_packages(self._read_lines_from_file(file_name))

    @uses_fabric
    def update_apt_sources(self):
        """Update apt source."""
        logger.info("Update apt source.")
        with settings(warn_only=True):
            sudo(
                "export DEBIAN_FRONTEND=noninteractive ; apt-get -qq update || apt-get -qq update"
            )

    @uses_fabric
    def upgrade_packages(self):
        """Safe upgrade of all packages."""

        self.update_apt_sources()
        # make sure apt/dpkg keep our installed config files, if any, and don't
        # prompt for user input:
        logger.info("Upgrading Packages")
        sudo(
            "export DEBIAN_FRONTEND=noninteractive ; apt-get dist-upgrade -y "
            '-o Dpkg::Options::="--force-confdef" '
            '-o Dpkg::Options::="--force-confold" --force-yes'
        )

    @uses_fabric
    def add_ppa(self, name):
        """Add personal package archive."""

        if self.ubuntu_release >= Decimal("12.04"):
            sudo("apt-add-repository -y %s" % name)
        else:
            sudo("apt-add-repository %s" % name)
        self.update_apt_sources()

    @uses_fabric
    def add_aptrepo(self, url, dist, repo_name, key_name=None, key_server=None):
        repo = " ".join(["deb", url, dist, repo_name])
        files.append("/etc/apt/sources.list", repo, use_sudo=True)
        if key_server is None:
            key_server = "keyserver.ubuntu.com"
        if key_name:
            sudo("apt-key adv --keyserver {0} --recv {1}".format(key_server, key_name))
        self.update_apt_sources()

    @uses_fabric
    def add_ppas_from_file(self, file_name):
        """Add personal package archive from a file list."""

        for ppa in self._read_lines_from_file(file_name):
            self.add_ppa(ppa)


class AptMixinMetaclass(type):
    """
    Metaclass that grabs all ``attr_prefix`` attributes from base classes and
    merges them into the ``_package_names`` attribute.  Allows package mixins
    to be defined in a declarative syntax.
    """

    def __new__(cls, name, bases, attrs):
        new_class = super(AptMixinMetaclass, cls).__new__(cls, name, bases, attrs)
        new_class._package_names = set(
            [b.package_name for b in bases if hasattr(b, "package_name")]
        )
        if "package_name" in attrs:
            new_class._package_names.add(attrs["package_name"])
        return new_class


class AptMixin(BaseAptMixin, metaclass=AptMixinMetaclass):
    def setup(self, propagate=True):
        super(AptMixin, self).setup()
        for attr_prefix in self._package_names:
            ppa = getattr(self, "%s_ppa" % attr_prefix, None)
            aptrepo = getattr(self, "%s_aptrepo" % attr_prefix, None)
            packages = getattr(self, "%s_packages" % attr_prefix, [])
            if ppa:
                self.add_ppa(ppa)
            if aptrepo:
                self.add_aptrepo(*aptrepo)
            self.install_packages(packages)



from fabric.api import sudo

from fabulaws.decorators import uses_fabric
from fabulaws.ubuntu.packages.base import AptMixin


class MongoDbMixin(AptMixin):
    """
    FabulAWS Ubuntu mixin that installs and configures MongoDB from the
    distro-supplied apt package ``mongodb-server``.
    """

    package_name = "mongodb"
    mongodb_packages = ["mongodb-server"]

    @uses_fabric
    def mongodb_service(self, cmd):
        sudo("service mongodb {0}".format(cmd))

    def bind_app_directories(self, *args, **kwargs):
        # make sure we stop first in case we're being moved to a secure directory
        self.mongodb_service("stop")
        super(MongoDbMixin, self).bind_app_directories(*args, **kwargs)
        self.mongodb_service("start")


class MongoDb10genMixin(MongoDbMixin):
    """
    FabulAWS Ubuntu mixin that installs and configures MongoDB from the 10gen
    apt repository, which tends to be more up to date.
    """

    mongodb_aptrepo = (
        "http://downloads-distro.mongodb.org/repo/ubuntu-upstart",
        "dist",
        "10gen",
        "7F0CEB10",
    )
    mongodb_packages = ["mongodb-10gen"]



from setuptools import find_packages, setup

setup(
    name="fabulaws",
    version=__import__("fabulaws").__version__,
    author="Caktus Consulting Group",
    author_email="solutions@caktusgroup.com",
    packages=find_packages(),
    include_package_data=True,
    url="http://github.com/caktus/fabulaws/",
    license="BSD",
    description="Simple tool for interacting with AWS in Python",
    classifiers=[
        "Topic :: System :: Systems Administration",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
    ],
    long_description=open("README.rst").read(),
    install_requires=[
        "pyyaml<5.5,>=3.10",
        "boto>=2.39,<3",
        "fabric<2.0",
        "paramiko==2.11.0",
    ],
)



import copy
import logging
import os
import time
from decimal import Decimal
from getpass import getpass

import boto.exception
from fabric.api import hide, put, run, sudo
from fabric.contrib import files

from fabulaws.api import answer_sudo, call_python
from fabulaws.decorators import uses_fabric
from fabulaws.ec2 import EC2Instance
from fabulaws.ubuntu.packages.base import BaseAptMixin

__all__ = ["UbuntuInstance"]

logger = logging.getLogger("fabulaws.ubuntu.instances.base")


class UbuntuInstance(BaseAptMixin, EC2Instance):
    """
    Base class for all Ubuntu instances.
    """

    user = "ubuntu"
    admin_groups = ["admin"]
    volume_info = []  # tuples of (device, mount_point, size_in_GB, type, passwd)
    fs_type = "ext3"
    fs_encrypt = False
    ebs_encrypt = False
    ubuntu_mirror = None
    mount_script = "/mount-deferred.sh"

    def __init__(self, *args, **kwargs):
        self.volumes = []
        super(UbuntuInstance, self).__init__(*args, **kwargs)

    @uses_fabric
    def _get_nvme_devs(self):
        """
        Attempts to identify any nvme devices per:
        https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/nvme-ebs-volumes.html
        These are typically used on newer instance types (t3+, m5+, etc.).
        """
        nvme_devs = {}
        for instance_dev in sudo("lsblk|grep '^nvme'|cut -d' ' -f 1").split("\n"):
            instance_dev = instance_dev.strip()  # remove "\r" if it exists
            devmap_dev = sudo(
                "nvme id-ctrl -v /dev/%s|grep '^0000:'|cut -d'\"' -f2|tr -d '.'"
                % instance_dev
            )
            nvme_devs[devmap_dev] = "/dev/%s" % instance_dev
        return nvme_devs

    @uses_fabric
    def _wait_for_device(self, device, max_tries=30):
        """
        Waits for the given device to manifest in /dev, and returns the actual
        path of the device.
        """
        # try all possible paths for the device in turn until one shows up
        devices = [device, device.replace("/dev/sd", "/dev/xvd")]
        logger.info("Waiting for device {0} to appear".format(device))
        self.install_packages(["nvme-cli"])
        for _ in range(max_tries):
            nvme_devs = self._get_nvme_devs()
            for device in devices:
                if device in nvme_devs:
                    device = nvme_devs[device]
                if files.exists(device):
                    logger.debug("Found device {0}".format(device))
                    return device
            time.sleep(1)
        # Device wasn't found; run dmesg to hopefully tell us where it is
        run("dmesg")
        raise ValueError(
            "Devices {0} never appeared; nvme_devs={1}.".format(
                ", ".join(devices),
                str(nvme_devs),
            )
        )

    @uses_fabric
    def _encrypt_device(self, device, passwd=None):
        """
        Encrypts the given device.  If a password is not specified,
        a prompt will be issued.
        """
        if not passwd:
            passwd = getpass.getpass("Enter LUKS passphrase for cryptsetup: ")
        self.install_packages(["python3-pexpect", "cryptsetup"])
        crypt = "crypt-{0}".format(device.split("/")[-1])
        with hide("stdout"):
            answers = [
                (r"Are you sure\? \(Type uppercase yes\):", "YES"),
                ("Enter .*passphrase:", passwd),
                ("Verify passphrase:", passwd),
            ]
            answer_sudo(
                "cryptsetup -y luksFormat {device}".format(device=device),
                answers=answers,
            )
            answers = [("Enter passphrase for .+:", passwd)]
            answer_sudo(
                "cryptsetup luksOpen {device} {crypt}"
                "".format(device=device, crypt=crypt),
                answers=answers,
            )
        device = "/dev/mapper/{0}".format(crypt)
        return device

    @uses_fabric
    def _mount_and_persist(self, device, mount_point, fs_type=None, opts=None):
        mount_cmd = " ".join(
            [
                "mount",
                "-o {}".format(opts) if opts else "",
                "-t {}".format(fs_type) if fs_type else "",
                device,
                mount_point,
            ]
        )
        fstab_entry = " ".join(
            [
                device,
                mount_point,
                fs_type if fs_type else self.fs_type,
                opts if opts else "defaults",
                "0",
                "0",
            ]
        )
        sudo(mount_cmd)
        if self.fs_encrypt:
            if not files.exists(self.mount_script):
                files.append(self.mount_script, "#!/bin/sh", use_sudo=True)
                sudo("chmod a+x {}".format(self.mount_script))
            files.append(self.mount_script, mount_cmd, use_sudo=True)
        else:
            files.append("/etc/fstab", fstab_entry, use_sudo=True)

    @uses_fabric
    def _format_volume(self, device, mount_point, passwd=None):
        """
        Creates an EBS volume of size ``vol_size``, manifests the device at
        ``device`` in this instance, and mounts it at ``mount_point``.
        """
        logger.info("Formatting {0}".format(device))
        if self.fs_encrypt:
            device = self._encrypt_device(device, passwd)
        sudo("mkfs.{0} {1}".format(self.fs_type, device))
        sudo("mkdir {0}".format(mount_point))
        self._mount_and_persist(device, mount_point)

    def _set_volume_tags(self, vol, device, tags=None):
        if tags is None:
            tags = copy.copy(self._tags) or {}
        if "Name" in tags:
            tags["Name"] = "_".join([tags["Name"], os.path.basename(device)])
        else:
            tags["Name"] = "_".join([self.instance.id, os.path.basename(device)])
        tags["device"] = device
        self.conn.create_tags([vol.id], tags)

    def add_tags(self, tags):
        """
        Propagate any tag updates on the instance to the associated volumes
        (e.g., for name changes).
        """
        super(UbuntuInstance, self).add_tags(tags)
        # allow super class to update self._tags and use that in _set_volume_tags
        for vol, (device, _, _, _, _) in zip(self.volumes, self.volume_info):
            self._set_volume_tags(vol, device)

    @uses_fabric
    def _create_volume(self, device, mount_point, vol_size, vol_type, passwd=None):
        """
        Creates an EBS volume of size ``vol_size``, manifests the device at
        ``device`` in this instance, and mounts it at ``mount_point``.
        """
        logger.info("Attaching {0}".format(device))
        inst = self.instance
        # the placement is the availability zone
        vol = self.conn.create_volume(
            vol_size, inst.placement, volume_type=vol_type, encrypted=self.ebs_encrypt
        )
        self._set_volume_tags(vol, device)
        try:
            logger.debug(
                "Waiting for volume {0} to become AVAILABLE "
                "(current state={1})".format(vol.id, vol.volume_state())
            )
            while vol.volume_state() != "available":
                time.sleep(1)
                vol.update()
                logger.debug("  current state: {0}".format(vol.volume_state()))
            vol.attach(inst.id, device)
            logger.debug(
                "Waiting for volume {0} to become ATTACHED "
                "(current state={1})".format(vol.id, vol.volume_state())
            )
            while vol.volume_state() != "in-use":
                time.sleep(1)
                vol.update()
                logger.debug("  current state: {0}".format(vol.volume_state()))
        except:  # noqa: E722
            self._destroy_volume(vol)
            raise
        return vol

    def _destroy_volume(self, vol):
        """
        Forcibly detaches and destroys the given EBS volume, where vol is an
        instance of ``boto.ec2.volume``.
        """
        logger.debug("Detaching volume {0}".format(vol.id))
        try:
            vol.detach(force=True)
        except boto.exception.EC2ResponseError:
            logger.exception("Failed to detach volume; continuing anyway.")
        logger.debug("Waiting for volume {0} to become " "available".format(vol.id))
        while vol.volume_state() == "in-use":
            time.sleep(1)
            vol.update()
        logger.debug("Deleting volume {0}".format(vol.id))
        vol.delete()

    @property
    @uses_fabric
    def server_memory(self):
        """Returns total server memory, in MB"""
        if not hasattr(self, "_server_memory"):
            mem = run("cat /proc/meminfo|grep MemTotal")
            self._server_memory = int(mem.split()[1]) / 1024
        return self._server_memory

    @property
    @uses_fabric
    def ubuntu_release(self):
        if not hasattr(self, "_ubuntu_release"):
            self._ubuntu_release = Decimal(run("lsb_release -r -s").strip())
        return self._ubuntu_release

    @uses_fabric
    def setup_mirror(self, mirror=None):
        if not mirror:
            mirror = self.ubuntu_mirror
        if mirror:
            orig = "{region}.ec2.archive.ubuntu.com".format(region=self.region)
            mirror = mirror.format(region=self.region)
            files.sed("/etc/apt/sources.list", orig, mirror, use_sudo=True)

    def setup(self):
        """
        Extends the base EC2Instance ``setup()`` method with routines to
        update apt sources on the instance.  Also creates volumes defined in
        the ``volume_info`` list on this instance.
        """
        super(UbuntuInstance, self).setup()
        # this is required because we may need to install cryptsetup when
        # creating volumes
        self.setup_mirror()
        self.update_apt_sources()
        # the first apt-get update may update sources.list, so re-run it here
        self.setup_mirror()
        self.update_apt_sources()
        for vol in self.volume_info:
            if len(vol) == 5:
                device, mount_point, vol_size, vol_type, passwd = vol
            else:
                raise Exception(
                    "volume_info must be populated with tuples of "
                    "(device, mount_point, vol_size, vol_type, passwd)"
                )
            if device == "instance-store":
                with self:
                    device = run("mount|grep /mnt|cut -d' ' -f1").strip()
                    sudo("umount {0}".format(device))
            else:
                self.volumes.append(
                    self._create_volume(device, mount_point, vol_size, vol_type, passwd)
                )
            device = self._wait_for_device(device)
            self._format_volume(device, mount_point, passwd)

    @uses_fabric
    def create_users(self, users, ignore_existing=True):
        """
        Create admin users and deploy SSH keys to the server.  ``users`` is
        a list of (username, keyfile) tuples.  The users will be created with
        empty passwords.
        """

        # This reverts Ubuntu >= 21.04 back to more permissive home directory permissions
        # https://ubuntu.com/blog/private-home-directories-for-ubuntu-21-04
        # Without this many setup processes fail because they cannot access files and directories
        # within the /home/sam directory.
        sudo("dpkg-reconfigure -f noninteractive adduser")
        sudo(r"sed -i 's/^\(HOME_MODE\s\+0750\)/#\1/' /etc/login.defs")

        if self.admin_groups:
            groups = "-G {0}".format(",".join(self.admin_groups))
        else:
            groups = ""
        for name, keyfile in users:
            if ignore_existing and files.exists("/home/{0}".format(name)):
                logger.info("Not creating existing user {0}".format(name))
            else:
                sudo("useradd -m {0} -s /bin/bash {1}".format(groups, name))
                sudo("passwd -d {0}".format(name))
                sudo("mkdir /home/{0}/.ssh".format(name))
            put(
                keyfile,
                "/home/{0}/.ssh/authorized_keys2".format(name),
                use_sudo=True,
                mode=0o600,
            )
            sudo("chown -R {0} /home/{0}/.ssh".format(name))
            # if a file exists with the comment field (e.g., for GECOS info)
            # for the user, use usermod -c to add it.
            gecos_file = keyfile + ".gecos"
            if os.path.exists(gecos_file):
                with open(gecos_file, "r") as gecos_fd:
                    gecos = gecos_fd.readline().strip()
                    sudo('usermod -c "{}" {}'.format(gecos, name))

    @uses_fabric
    def bind_app_directories(self, app_dirs, app_root):
        """
        Move the given directories, ``app_dirs'', to the specified file system
        mounted at ``app_root'' (optionally secure if ``self.fs_encrypt == True'').
        """
        assert files.exists(app_root)
        for app_dir in app_dirs:
            bound_app_dir = "".join([app_root, app_dir])
            bound_parent_dir = call_python("os.path.dirname", bound_app_dir)
            if files.exists(app_dir):
                sudo("mkdir -p {0}".format(bound_parent_dir))
                sudo("mv {0} {1}".format(app_dir, bound_app_dir))
            else:
                sudo("mkdir -p {0}".format(bound_app_dir))
            sudo("mkdir -p {0}".format(app_dir))
            self._mount_and_persist(bound_app_dir, app_dir, fs_type="none", opts="bind")

    def cleanup(self):
        """
        If needed, destroys any volumes created for this instance and then
        calls the base class's ``cleanup()`` method.
        """
        if self._terminate:
            while self.volumes:
                vol = self.volumes.pop()
                self._destroy_volume(vol)
        else:
            for vol in self.volumes:
                logger.warning(
                    'Left volume "{0}" in state "{1}"'
                    "".format(vol.id, vol.volume_state())
                )
        super(UbuntuInstance, self).cleanup()
